// Scintilla source code edit control
// ScintillaTK.cxx - Tk specific subclass of ScintillaBase
// Copyright 1998-2004 by Neil Hodgson <neilh@scintilla.org>
// The License.txt file describes the conditions under which this software may be
// distributed.
// Copyright (c) 2013 Mentor Graphics Corporation
// The license.terms file describes the conditions under which this software may be distributed.

#include <tk.h>
#include <stdlib.h>
#include <string>
#include <string.h>
#include <vector>
#include <map>

#include "Platform.h"
#include "SplitVector.h"

#include "ILexer.h"
#include "Scintilla.h"
#include "ScintillaTk.h"

#include "AutoComplete.h"
#include "Partitioning.h"
#include "CallTip.h"
#include "CellBuffer.h"
#include "CharClassify.h"
#include "RunStyles.h"
#include "ContractionState.h"
#include "Decoration.h"
#include "Document.h"
#include "KeyMap.h"
#include "Indicator.h"
#include "XPM.h"
#include "LineMarker.h"
#include "Style.h"
#include "ViewStyle.h"
#include "Selection.h"
#include "PositionCache.h"
#include "Editor.h"
#include "ScintillaBase.h"

#ifdef SCI_LEXER
#include "SciLexer.h"
#include "LexerModule.h"
#include "ExternalLexer.h"
#endif

#include "scintilla-ext.h"
using namespace TkSciExt;

#ifdef SCI_NAMESPACE
using namespace Scintilla;
#endif



/////////////////////////////////////////////////////////////////////
// CLASS: ScintillaTK
//
/////////////////////////////////////////////////////////////////////
class ScintillaTK : public ScintillaBase {

   // Private so ScintillaTK objects can not be copied
   ScintillaTK(const ScintillaTK &);
   ScintillaTK &operator=(const ScintillaTK &);

public:
   ScintillaTK(_ScintillaObject *sci_);
   virtual ~ScintillaTK();

private:
   // Required virtuals (ScintillaBase)
   virtual void Initialise();
   virtual void Finalise();

   // Required virtuals (Editor)
   virtual sptr_t DefWndProc(unsigned int iMessage, uptr_t wParam, sptr_t lParam);
   virtual void ClaimSelection();
   virtual void Copy();
   virtual void CopyToClipboard(const SelectionText &selectedText);
   virtual void Paste();
   virtual void SetVerticalScrollPos();
   virtual void SetHorizontalScrollPos();
   virtual bool ModifyScrollBars(int nMax, int nPage);
   virtual void NotifyChange();
   virtual void NotifyParent(SCNotification scn);
   virtual void SetTicking(bool on);
   virtual void SetMouseCapture(bool on);
   virtual bool HaveMouseCapture();
   virtual void CreateCallTipWindow(PRectangle rc) {}
   virtual void AddToPopUp(const char *label, int cmd = 0, bool enabled = true) {}

public:  // Public for client code to call back to scintilla
   virtual sptr_t WndProc(unsigned int iMessage, uptr_t wParam, sptr_t lParam);

   // public to allow static "timeout" function access
   void Tick();
private:
   void CopyToScreen(bool margin_only = false);
   void PaintRect(PRectangle r);
   void FullPaint();

   void StoreOnClipboard(SelectionText *clipText);
   virtual PRectangle GetClientRectangle();
   virtual void DisplayCursor(Cursor c); // overload
   virtual void ScrollText(int linesToMove);

   bool capturedMouse;
   ScintillaObject *sc_obj;
   Surface *surfaceWindow;
   int press_button; // used to indicate which button was pressed
   int press_x;      // records X location of the press
   int press_y;      // Y location
   bool		  painted_region_set;
   PRectangle PaintedRegion;
};

//vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
// declared in ScintillaTk.h
#ifdef _WIN32
void _print_trace (bool on) {}
#else
#include <execinfo.h> //FOR DUMPING THE CALL STACK...
void _print_trace (bool on) {
if (!on)
return;
#define TLEVELS 20
   void *array[TLEVELS];
   size_t size, i;
   char **strings;
   size = backtrace (array, TLEVELS);
   strings = backtrace_symbols (array, size);
   printf ("Obtained %d stack frames.\n", (int)size);
   for (i = 0; i < size; i++) printf (">>> %s\n", strings[i]);
   free (strings);
}
#endif
static int _x_error_count = 0;
int _x_error_handler(Display *dsp, XErrorEvent *evt) {
#ifndef _WIN32
   char buf[1000];
   if (_x_error_count++ > 10) {
      printf("X ERROR LIMIT EXCEEDED ... Bailing!!\n");
      exit(99);
      return 0;
   }
   XGetErrorText( dsp, evt->error_code, buf, 1000);
   printf("XERROR: %s\n\tDisplay=%p type=%d resourceid=%ld serial=%ld\n\terror_code=%d request_code=%d minor_code=%d\n",
      buf,
      (void *)dsp, evt->type, evt->resourceid, evt->serial, 
      evt->error_code, evt->request_code, evt->minor_code);
   _print_trace(1);
#endif
   return 0;
}
//^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

void ScintillaTK::Initialise() {
   //TODO: what goes in here???
printf("ScintillaTK::Initialise\n");
// maybe need to set up blinking of caret?
}
void ScintillaTK::Finalise() {
//   SetTicking(false);
   ScintillaBase::Finalise();
printf("ScintillaTK::Finalise\n");
}

//=============================================================
//CONSTRUCTOR
//=============================================================
extern int _Xdebug;

ScintillaTK::ScintillaTK(ScintillaObject *sc_obj_) :
   capturedMouse(false),
   surfaceWindow(0)
{
//printf("ScintillaTK CTOR this=%p\n",(void *)this);
   sc_obj = sc_obj_;
   wMain = sc_obj; // base class relies on this being set!!
//printf("CTOR >>>>>>>> wMain=%p\n",(void *)sc_obj);
#if 0
   //NOTE: Don't use this when running within vish !!
   _Xdebug = 1; // <<< this is a global in the X11 code
   XSetErrorHandler(_x_error_handler);
#endif
}
ScintillaTK::~ScintillaTK() {
//printf("ScintillaTK DTOR this=%p\n",(void *)this);
   if (surfaceWindow) {
      surfaceWindow->Release();

      if (timer.ticking) {
         Tcl_DeleteTimerHandler(reinterpret_cast<Tcl_TimerToken>(timer.tickerID));
      }
      delete surfaceWindow;
   }
}


/****************************************************************
// Func: WndProc
// Desc: Handles message processing
*****************************************************************/
sptr_t ScintillaTK::WndProc(unsigned int iMessage, uptr_t wParam, sptr_t lParam) {
//printf("ScintillaTK::WndProc iMessage=%d\n",iMessage);
   try {
      switch (iMessage) {

         case SCI_TKMSG_ANNOTATIONLENGTH: {
            Document *doc = reinterpret_cast<Document *>(ScintillaBase::WndProc(SCI_GETDOCPOINTER,0,0));

            return doc->AnnotationStyledText(wParam).length;
         }
         case SCI_TKMSG_ANNOTATIONINDENT: {

            int line = static_cast<int>(wParam);
            bool pixels = static_cast<int>(lParam);

            Document *doc = reinterpret_cast<Document *>(ScintillaBase::WndProc(SCI_GETDOCPOINTER,0,0));

            // the indent value represents the number of spaces(tabs are converted to
            // the required number of spaces) preceding the text
            int indent = doc->GetLineIndentation(line);

            if ( ScintillaBase::WndProc(SCI_ANNOTATIONGETVISIBLE, 0,0) == ANNOTATION_BOXED ) {
               indent++; // box style adds extra space
            }
            if (pixels) {
               // adjust for width of margin(s) and pad region leftside of text
               int marginwidth = TkSciExt::GetMarginWidth((ClientData)sc_obj->client);
               marginwidth += ScintillaBase::WndProc(SCI_GETMARGINLEFT, 0,0);

               // adjust for hz scroll
               int xoff = ScintillaBase::WndProc(SCI_GETXOFFSET, 0,0);

               // the returned value will represent the X pixel location
               // of where the annotation text's left edge is.
               // Note: "vs" is a protected variable in Editor base class
               return (indent * vs.spaceWidth) + marginwidth - xoff;
            }
            return indent;
         }
		 case SCI_TKMSG_RESIZE : {
            ChangeSize();
			break;
         }
         case SCI_TKMSG_PAINT: {
            PRectangle *rc = reinterpret_cast<PRectangle *>(wParam);
            PaintRect(*rc);
            break;
         }
         case SCI_TKMSG_DESTROY: {
            delete this;
            break;
         }
         case SCI_TKMSG_DUMPSTYLES: {
            if (wParam != 0) {
               int n = static_cast<int>(ScintillaBase::WndProc(SCI_GETLENGTH,0,0));
               int last = -1;
               for (int i=0;i < n;i++) {
                  int curr = static_cast<int>(ScintillaBase::WndProc(SCI_GETSTYLEAT, i,0));
                  if (curr != last) {
                     last = curr;
                     printf("{%d}", curr); // << current style id
                  }
                  printf("%c", static_cast<char>(ScintillaBase::WndProc(SCI_GETCHARAT,i,0)));
               }
               printf("\n");
            }
            break;
         }
#ifdef SCI_LEXER
         case SCI_LOADLEXERLIBRARY:
            // support for dynamically loading external lexers
            LexerManager::GetInstance()->Load(reinterpret_cast<const char *>(lParam));
            break;
#endif
         case SCI_SETFOCUS: {
            //### KEEP THIS CASE ENTRY _LAST_ ###
            SetTicking(wParam != 0);
            // no "break", let base message handling deal with the rest
         }
         default:
            return ScintillaBase::WndProc(iMessage, wParam, lParam);
      }
   } catch (std::bad_alloc&) {
      errorStatus = SC_STATUS_BADALLOC;
   } catch (...) {
      errorStatus = SC_STATUS_FAILURE;
   }
   return 0l;
}
/****************************************************************
// Func: DefWndProc
// Desc: Required by Editor base class. It's called if the base
//       class's WndProc function doesn't handle a message
*****************************************************************/
sptr_t ScintillaTK::DefWndProc(unsigned int, uptr_t, sptr_t) {
   return 0;
}

/****************************************************************
// Func: PaintRect
// Desc: 
*****************************************************************/
void ScintillaTK::PaintRect(PRectangle r) {

   try {
      paintState = painting; // see Editor.h
      rcPaint = r;
      PRectangle rcClient = GetClientRectangle();

//printf("PaintRect: +%02d+%02dx%03dx%03d\n",rcPaint.left,rcPaint.top,(rcPaint.right-rcPaint.left),(rcPaint.bottom-rcPaint.top));

      paintingAllText = rcPaint.Contains(rcClient);
      if ( ! surfaceWindow) {
         surfaceWindow = Surface::Allocate(SC_TECHNOLOGY_DEFAULT);
      }
//printf("PaintRect: SURFACE=%p\n",(void *)surfaceWindow);
      if (surfaceWindow) {
         sc_obj->main = true;
         surfaceWindow->Init(sc_obj);
         painted_region_set = true;
         PaintedRegion = rcPaint;
         Paint(surfaceWindow, rcPaint); // call base class's Paint
         painted_region_set = false;

//printf("PaintRect: DONE\n^^^^^^^^^^\n");
      }
      if (paintState == paintAbandoned) {
         // Painting area was insufficient to cover new styling or brace highlight positions
         FullPaint();
      }
      paintState = notPainting;

   } catch (...) {
      // don't know how we would get here, but ScintiallaGTK does it this way...
      errorStatus = SC_STATUS_FAILURE;
   }
}

/****************************************************************
// Func: ScrollText
// Desc: Called from Editor::ScrollTo
*****************************************************************/
void ScintillaTK::ScrollText(int linesToMove) {

   int diff = vs.lineHeight * -linesToMove;
//printf("ScrollText %d pixels\n",diff);
   ClientData d = (ClientData)sc_obj->client;
   TkSciExt::SetYScrollDelta(d, diff);

   TkSciExt::UpdateForScroll(d);
//    if ( !(sciwidgetPtr->flags & REDRAW_PENDING) ) {
// //printf("ScrollText -- QUEUING DISPLAY\n");
//       Tk_DoWhenIdle(SciWidgetDisplay, (ClientData)sciwidgetPtr);

//    } else {
// //printf("ScrollText -- SKIPPING REDRAW...\n");
//    }
//    sciwidgetPtr->flags |= REDRAW_PENDING|REDRAW_YSCROLL;

//printf("OVERLOADED ScrollText moving %d lines\n",diff);
//   Editor::ScrollText(linesToMove);
}

/****************************************************************
// Func: GetClientRectangle
// Desc: Returns the bounding rectangle encompassing the parent
//       window
*****************************************************************/
PRectangle ScintillaTK::GetClientRectangle() {
   Tk_Window tkwin = sc_obj->tkwin;
   int w = Tk_Width(tkwin);
   int h = Tk_Height(tkwin);
//TODO: factor in borders???
int bd=0;
   PRectangle rc(bd,bd, w-(bd*2), h-(bd*2));
   return rc;
}

void ScintillaTK::DisplayCursor(Cursor c) {
printf("ScintillaTK::DisplayCursor\n");
}

// Redraw all of text area. This paint will not be abandoned.
void ScintillaTK::FullPaint() {
	paintState = painting;
	rcPaint = GetClientRectangle();
	paintingAllText = true;
	if (surfaceWindow) {
		Paint(surfaceWindow, rcPaint);
	} 
}

void ScintillaTK::Copy() {

   if (!sel.Empty()) {
      SelectionText *clipText = new SelectionText();
      CopySelectionRange(clipText);
      StoreOnClipboard(clipText);
   }
}
void ScintillaTK::CopyToClipboard(const SelectionText &selectedText) {

   SelectionText *clipText = new SelectionText();
   clipText->Copy(selectedText);
   StoreOnClipboard(clipText);
}
void ScintillaTK::StoreOnClipboard(SelectionText *clipText) {
//TODO: what do we do??
//Tk_ClipboardClear + Tk_ClipboardAppend ??
}
void ScintillaTK::Paste() {
//TODO: what do we do??
}

void ScintillaTK::ClaimSelection() {
//    // X Windows has a 'primary selection' as well as the clipboard.
//    // Whenever the user selects some text, we become the primary selection
//    if (!sel.Empty() && IS_WIDGET_REALIZED(GTK_WIDGET(PWidget(wMain)))) {
//       primarySelection = true;
//       gtk_selection_owner_set(GTK_WIDGET(PWidget(wMain)),
//                               GDK_SELECTION_PRIMARY, GDK_CURRENT_TIME);
//       primary.Free();
//    } else if (OwnPrimarySelection()) {
//       primarySelection = true;
//       if (primary.s == NULL)
//          gtk_selection_owner_set(NULL, GDK_SELECTION_PRIMARY, GDK_CURRENT_TIME);
//    } else {
//       primarySelection = false;
//       primary.Free();
//    }
}

// *required* virtual
void ScintillaTK::SetVerticalScrollPos() {
   DwellEnd(true);
   // most of the scroll functionality is handled in the widget's "yview" command

   TkSciExt::MSG_SyncYScroll((ClientData)sc_obj->client);
}

// *required* virtual
void ScintillaTK::SetHorizontalScrollPos() {
   DwellEnd(true);
   // the actual scroll functionality is handled in the widget's "xview" command

   TkSciExt::MSG_SyncXScroll((ClientData)sc_obj->client);
}

// *required* virtual
bool ScintillaTK::ModifyScrollBars(int nMax, int nPage) {
   bool modified = true;
   // this gets called _BEFORE_ the client window has it's size info set, which
   // means GetClientRectangle won't provide anything useful to the Editor code.
   // We'll deal with getting the initial scrollbar data set by handling the first
   // <Map> event - see widget.cxx in the SciWidgetEventProc function

   // CAVEAT: To support "scrollwidthtracking" we need to sync the horizontal
   // scrollbar because this function is called after Scintilla updates the
   // widest line info.
   if (!Tk_IsMapped(sc_obj->tkwin)) {
	   // if the window isn't mapped don't waste time updating the scrollbars
	   // the map notify event will force a scrollbar update
	   return 0;
   }
   if (WndProc(SCI_GETHSCROLLBAR, 0,0) && WndProc(SCI_GETSCROLLWIDTHTRACKING, 0,0)) {
       TkSciExt::MSG_SyncXScroll((ClientData)sc_obj->client);
   }
   if (WndProc(SCI_GETVSCROLLBAR, 0,0)) {
       TkSciExt::MSG_SyncYScroll((ClientData)sc_obj->client);
   }
   return modified;
}

void ScintillaTK::NotifyChange() {
//something changed in the document...
//printf("NotifyChange\n");
}
#define _RGB_(r,g,b) r | (g << 8) | (b << 16)
#include <time.h>

void ScintillaTK::NotifyParent(SCNotification scn) {
//printf("ScintillaTK::NotifyParent scn=%d  SURFACE=%p\n", scn.nmhdr.code,(void *)surfaceWindow);

   switch (scn.nmhdr.code) {

       case SCN_MODIFIED: {
          int mt = scn.modificationType;
          if (mt & (SC_MOD_INSERTTEXT|SC_MOD_DELETETEXT)) {
             ClientData d = (ClientData)sc_obj->client;
             if (scn.linesAdded) {
                // send lines added message
                int ln = WndProc(SCI_LINEFROMPOSITION, scn.position, 0);

                TkSciExt::MSG_LinesAdded(d, ln, scn.linesAdded);
             }
             TkSciExt::MSG_Modified(d); // send notification of document modification
          } 
          break;
       }
      case SCN_PAINTED: {
         // This is called at the end of the Paint function, _BUT_ this
         // only happens if part of the text area had to be painted.
         //
         CopyToScreen();
         break;
      }
      case SCN_UPDATEUI: {
         // Note that this is how SciTE does updates for the statusbar's
         // line and column indicator.

         TkSciExt::MSG_MoveUpdate((ClientData)sc_obj->client);
         break;
      }
   }
}

void ScintillaTK::CopyToScreen(bool margin_only) {
   PRectangle r;
   Point pt = Point(0,0);
   if (painted_region_set) { 
		r = PaintedRegion;
   } else { 
	    r = GetClientRectangle();
   }

   if (margin_only) {

      int marginwidth = TkSciExt::GetMarginWidth((ClientData)sc_obj->client);
      if (marginwidth <= 0) return;
      r.right = marginwidth;
   }
   // copy offscreen pixmap used by drawing operations
   // to the screen
   surfaceWindow->Copy(r, pt, *surfaceWindow);
}


//////////////////////////////////////////////////////
// BEGIN caret(aka "tick") management
//////////////////////////////////////////////////////
static void _TimeOut(ClientData cd) {

   ScintillaTK *sc = reinterpret_cast<ScintillaTK *>(cd);
   sc->Tick(); // call our overloaded function
}
void ScintillaTK::Tick() {

   // Mimic how Tk shows the insert caret(which is 1-pixel wide
   // for the first column and thicker for the others)
   int currpos = CurrentPosition();
   int col = WndProc(SCI_GETCOLUMN, currpos, 0);
   vs.caretWidth = (col == 0 ? 1 : 2);

   Editor::Tick();

   // schedule another timer...
   timer.tickerID = reinterpret_cast<TickerID>(
      Tcl_CreateTimerHandler(timer.tickSize, _TimeOut, (ClientData)this)
   );
}
// This is called from Editor
void ScintillaTK::SetTicking(bool on) {

   if (timer.ticking != on) {
      timer.ticking = on;
      if (timer.ticking) {
         timer.tickerID = reinterpret_cast<TickerID>(
            Tcl_CreateTimerHandler(timer.tickSize, _TimeOut, (ClientData)this)
         );
      } else {
         Tcl_DeleteTimerHandler(reinterpret_cast<Tcl_TimerToken>(timer.tickerID));
      }
   }
   timer.ticksToWait = caret.period;
}
//////////////////////////////////////////////////////

void ScintillaTK::SetMouseCapture(bool on) {

   if (mouseDownCaptures) {
      if (on) {
         Tk_Grab(sc_obj->interp, sc_obj->tkwin, 0/*is_global*/);
      } else {
         Tk_Ungrab(sc_obj->tkwin);
      }
   }
   capturedMouse = on;
}
bool ScintillaTK::HaveMouseCapture() {
   return capturedMouse;
}

//--- end of class ScintillaTK

//#############################################################################
//#
//# The following two functions, which are declared in ScintillaTk.h, are
//# how Tk-based widgets create and communicate with the Scintilla editor
//#
//#############################################################################
ScintillaObject* scintilla_new(Tcl_Interp *interp, Tk_Window tkwin, ClientData cd) {

   ScintillaObject *sci_obj = new ScintillaObject;
   sci_obj->interp = interp;
   sci_obj->tkwin  = tkwin;
   sci_obj->client = cd;
   sci_obj->main = false;
   sci_obj->pscin  = new ScintillaTK(sci_obj);
   return sci_obj;
}

sptr_t scintilla_send_message(ScintillaObject *sci,unsigned int iMessage, uptr_t wParam, sptr_t lParam) {

   ScintillaTK *psci = reinterpret_cast<ScintillaTK *>(sci->pscin);
   return psci->WndProc(iMessage, wParam, lParam);
}
